import warnings
from brukeropus.file.constants import TYPE_CODE_LABELS, CODE_3_ABR


__docformat__ = "google"


def get_type_code_label(pos_idx: int, val: int):
    '''Returns the type code label of a file block given the position index and value of the type code.
    
    **Depreciation Warning:** function will be removed soon. Functionality has been integrated into `BlockType` class.

    The file blocks on an OPUS file feature six-integer type codes, for example (3, 1, 1, 2, 0, 0), that categorize the
    contents of the file block. The positional index defines the category, while the value at that index defines the
    specific type of that category.  For example, the first integer (pos_idx=0), describes the type of data in the
    block, if applicable:

        0: Undefined or N/A,
        1: Real Part of Complex Data,
        2: Imaginary Part of Complex Data,
        3: Amplitude

    This package includes the majority of type codes that OPUS uses, but in the event a type code label is not known,
    this function will return: "Unknown 0 4" where the first number is the position index, and the second is the
    unknown value integer.

    Args:
        pos_idx: positional index of the type code (0 - 5)
        val: value of the type code

    Returns:
        label (str): human-readable string label that describes the type code.
    '''
    warnings.warn('Depreciation warning: get_type_code_label will soon be removed. This functionality has moved to `BlockType` class')
    try:
        return TYPE_CODE_LABELS[pos_idx][val]
    except KeyError:
        return 'Unknown ' + str(pos_idx) + ' ' + str(val)


def get_block_type_label(block_type: tuple):
    '''Converts a six-integer tuple block type into a human readable label.
    
    **Depreciation Warning:** function will be removed soon. Functionality has been integrated into `BlockType` class.

    Args:
        block_type: six integer tuple found in the OPUS file directory that describes the block type

    Returns:
        label (str): human-readable string label
    '''
    warnings.warn('Depreciation warning: get_block_type_label will soon be removed. This functionality has moved to `BlockType` class')
    labels = [get_type_code_label(idx, val) for idx, val in enumerate(block_type) if val > 0
              and get_type_code_label(idx, val) != '']
    return ' '.join(labels)


def get_data_key(block_type: tuple):
    '''Returns a shorthand key for a given data block type: sm, rf, igsm, a, t, r, etc.

    Determines if the data block type is an interferogram, single-channel, absorption, etc. and whether it is associated
    with the sample or reference channel and returns a shortand key-like label: sm, rf, igsm, igrf, a, t, r, etc.  For
    the full data label (e.g. Sample Spectrum, Absorbance) use: get_block_type_label.
    This package includes the majority of type codes that OPUS uses, but in the event a type code label is not known,
    this function will return: "_33" or "sm_33" where 33 will change to the unkown block_type integer value.

    Args:
        block_type: six integer tuple found in the OPUS file directory that describes the block type

    Returns:
        key (str): shorthand string label that can be utilized as a data key (e.g. "sm", "igrf", "a")'''
    warnings.warn('Depreciation warning: get_data_key will soon be removed. This functionality has moved to `FileBlock` class')
    if block_type[3] in CODE_3_ABR.keys():
        key = CODE_3_ABR[block_type[3]]
        if block_type[1] == 1:
            key = merge_key(key, 'sm')
        elif block_type[1] == 2:
            key = merge_key(key, 'rf')
        elif block_type[1] > 3:
            key = key + '_' + str(block_type[1])
    else:
        key = '_' + str(block_type[3])
        if block_type[1] == 1:
            key = 'sm' + key
        elif block_type[1] == 2:
            key = 'rf' + key
        elif block_type[1] > 3:
            key = '_' + str(block_type[1]) + key
    if block_type[5] == 4:
        key = key + '_c'
    return key


def merge_key(key: str, sm: str):
    '''Merges "sm" or "rf" into an abreviated data key.  For special cases like ig or pw, the addition is appended
    (e.g. igsm, phrf), but for other cases, the addition is prepended (e.g. sm_2ch, rf_3ch)'''
    warnings.warn('Depreciation warning: merge_key will soon be removed. This functionality has moved to `FileBlock` class')
    if key[:2] in ['ig', 'ph', 'pw']:
        return key[:2] + sm + key[2:]
    else:
        return sm + key


import re, os
from brukeropus.file.block import FileBlock, BlockType
from brukeropus.file.parse import read_opus_file_bytes, parse_header, parse_directory
from brukeropus.file.params import get_param_label


__all__ = ['find_opus_files', 'parse_file_and_print']


__docformat__ = "google"


def find_opus_files(directory, recursive: bool = False):
    '''Finds all files in a directory with a strictly numeric extension (OPUS file convention).

    Returns a list of all files in directory that end in .# (e.g. file.0, file.1, file.1001, etc.). Setting recursive
    to true will search directory and all sub directories recursively. No attempt is made to verify the files are
    actually OPUS files (requires opening the file); the function simply looks for files that match the naming pattern.

    Args:
        directory (str or Path): path indicating directory to search
        recursive: Set to True to recursively search sub directories as well

    Returns:
        filepaths (list): list of filepaths that match OPUS naming convention (numeric extension)
    '''
    pattern = re.compile(r'.+\.[0-9]+$')
    file_list = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in filenames:
            if pattern.match(filename):
                file_list.append(os.path.join(root, filename))
        if not recursive:
            break
    return file_list


def parse_file_and_print(filepath, width=120):
    '''Parses an OPUS file and prints the block information as it goes along to the console.

    This function demonstrates the basic usage and interaction of the parsing functions.  It
    can also be used to diagnose a file parsing issue if one comes up.

    Args:
        filepath (str or Path): filepath to an OPUS file.
    '''
    filebytes = read_opus_file_bytes(filepath)
    if filebytes is not None:
        width = 120
        info_col_widths = (19, 15, 16, 61)
        info_col_labels = ('Block Type', 'Size (bytes)', 'Start (bytes)', 'Friendly Name')
        _print_block_header(filepath, width)
        version, dir_start, max_blocks, num_blocks = parse_header(filebytes)
        h_text = '    '.join([
            'Version: ' + str(version),
            'Directory start: ' + str(dir_start),
            'Max Blocks: ' + str(max_blocks),
            'Num Blocks: ' + str(num_blocks)])
        _print_centered(h_text, width)
        _print_block_header('Directory', width)
        block_infos = []
        _print_cols(info_col_labels, info_col_widths)
        for b_type, size, start in parse_directory(filebytes[dir_start:dir_start + num_blocks * 3 * 4]):
            try:
                b_type = BlockType(b_type)
                vals = [b_type.get_aligned_tuple_str(), size, start, b_type.label]
                _print_cols(vals, info_col_widths)
                block_infos.append((b_type, size, start))
            except Exception as e:
                print('Exception parsing block: ', e)
        for b_type, size, start in block_infos:
            try:
                block = FileBlock(filebytes, block_type=b_type, size=size, start=start)
                block.parse()
                _print_block(block, width=width)
            except Exception as e:
                print('Exception parsing block:', block.get_label(), '\n\tException:', e)
    else:
        print('Selected file is not an OPUS file: ', filepath)


def _print_block_header(label, width, sep='='):
    'Helper function for: parse_file_and_print'
    print('\n' + sep * width)
    _print_centered(label, width)


def _print_centered(text, width):
    'Helper function for: parse_file_and_print'
    print(' ' * int((width - len(text)) / 2) + text)


def _print_block(block: FileBlock, width: int):
    'Helper function for: parse_file_and_print'
    param_col_widths = (10, 45, 45)
    key_width = 10
    key_label_width = 45
    param_col_widths = (key_width, key_label_width, width - key_width - key_label_width)
    param_col_labels = ('Key', 'Friendly Name', 'Value')
    if not block.is_directory():
        _print_block_header(block.type.label, width)
        if block.is_param():
            _print_cols(param_col_labels, param_col_widths)
            if type(block.data) is dict:
                for key, val in block.data.items():
                    _print_cols((key.upper(), get_param_label(key), val), param_col_widths)
            else:
                print(block.data)
        elif block.is_data():
            print(block.data)
        elif block.is_data_series():
            if type(block.data) is dict:
                _print_centered('Num Blocks: ' + str(block.data['num_blocks']), width)
                _print_centered('Store Table: ' + str(block.data['store_table']), width)
                print(block.data['y'])
            else:
                print(block.data)
        elif block.is_file_log():
            print(block.data)
        else:
            _print_centered('Undefined Block Type: ' + str(block.type) + ' [Raw Bytes]', width)
            print(block.bytes)


def _print_cols(vals, col_widths,):
    'Helper function for: parse_file_and_print'
    string = ''
    for i, val in enumerate(vals):
        col_width = col_widths[i]
        val = str(val)
        if len(val) <= col_width - 2:
            string = string + val + ' ' * (col_width - len(val))
        else:
            string = string + val[:col_width - 5] + '...  '
    print(string)



from brukeropus.file.block import FileBlock
from brukeropus.file.parse import (parse_header,
                                   parse_directory,
                                   )


__docformat__ = "google"


class FileDirectory:
    '''Contains type and pointer information for all blocks of data in an OPUS file.

    `FileDirectory` information is decoded from the raw file bytes of an OPUS file. First the header is read which
    provides the start location of the directory block, number of blocks in file, and maximum number of blocks the file
    supports. Then it decodes the block pointer information from each entry of the file's directory block to create a
    `FileBlock` instance, initiates the block parsing, and adds the parsed block to the `blocks` attribute.

    Args:
        filebytes: raw bytes from OPUS file. see: `brukeropus.file.parser.read_opus_file_bytes`
        debug: whether to read the file in debug mode (default: False)

    Attributes:
        start: pointer to start location of the directory block
        max_blocks: maximum number of blocks supported by file
        num_blocks: total number of blocks in the file
        blocks: list of `FileBlock` from the file. The class parses these blocks upon initilization of the class.
        parse_error_blocks: list of `FileBlock` that encountered an error when attempting to parse the binary data.
    '''
    def __init__(self, filebytes: bytes, debug: bool = False):
        self.version, self.start, self.max_blocks, self.num_blocks = parse_header(filebytes)
        size = self.max_blocks * 3 * 4
        self.blocks = []
        self.parse_error_blocks = []
        self.toc = []
        for block_type, size, start in parse_directory(filebytes[self.start: self.start + size]):
            block = FileBlock(filebytes=filebytes, block_type=block_type, size=size, start=start, debug=debug)
            block.parse()
            if block.parse_error:
                self.parse_error_blocks.append(block)
            else:
                self.blocks.append(block)
    
    def __str__(self):
        if len(self.toc) == 0:
            return 'Blank OPUS file directory'
        else:
            label_width = max([len(e['type'].get_label()) for e in self.toc])
            type_width = 13
            start_width = 7
            size_width = 6
            attr_width = max([len(e['attr']) for e in self.toc])
            info = '    '.join(
                (f'Version: {self.version}',
                 f'Num Blocks: {len(self.toc)}/{self.num_blocks}[{self.max_blocks}]',
                 f'Start: {self.start}'))
            space = '   '
            header = space.join(
                (f'{"File Block Name":<{label_width}}',
                 f'{"Type Code":^{type_width}}',
                 f'{"Start":>{start_width}}',
                 f'{"Size":>{size_width}}',
                 f'{"Attribute":<{attr_width}}'))
            lines = [f'{"  Directory  ":=^{len(header)}}',
                     f'{info:^{len(header)}}',
                     '―'*len(header),
                     header,
                     '―'*len(header)]
            for entry in self.toc:
                string = space.join(
                    (f'{entry["type"].get_label():<{label_width}}',
                     f'{entry["type"].get_aligned_tuple_str():^{type_width}}',
                     f'{entry["start"]:>{start_width}}',
                     f'{entry["size"]:>{size_width}}',
                     f'{entry["attr"]:<{attr_width}}'))
                lines.append(string)
            return '\n'.join(lines)


